from PyQt5.QtCore import (
    Qt,
    QRect,
    QSize,
    QPoint,
)
from PyQt5.QtWidgets import (
    QLayout,
    QStyle,
    QSizePolicy,
)


class FlowLayout(QLayout):
    def __init__(self, parent=None, margin=-1, hspacing=-1, vspacing=-1):
        super(FlowLayout, self).__init__(parent)
        self._hspacing = hspacing
        self._vspacing = vspacing
        self._items = []
        self.setContentsMargins(margin, margin, margin, margin)
        self.setObjectName(type(self).__name__)

    def __del__(self):
        del self._items[:]

    def addItem(self, item):
        self._items.append(item)

    def horizontalSpacing(self):
        if self._hspacing >= 0:
            return self._hspacing
        else:
            return self.smartSpacing(QStyle.PM_LayoutHorizontalSpacing)

    def verticalSpacing(self):
        if self._vspacing >= 0:
            return self._vspacing
        else:
            return self.smartSpacing(QStyle.PM_LayoutVerticalSpacing)

    def count(self):
        return len(self._items)

    def itemAt(self, index):
        if 0 <= index < len(self._items):
            return self._items[index]

    def takeAt(self, index):
        if 0 <= index < len(self._items):
            return self._items.pop(index)

    def expandingDirections(self):
        return Qt.Orientations(0)

    def hasHeightForWidth(self):
        return True

    def heightForWidth(self, width):
        return self.doLayout(QRect(0, 0, width, 0), True)

    def setGeometry(self, rect):
        super(FlowLayout, self).setGeometry(rect)
        self.doLayout(rect, False)

    def sizeHint(self):
        return self.minimumSize()

    def minimumSize(self):
        size = QSize()
        for item in self._items:
            size = size.expandedTo(item.minimumSize())
        left, top, right, bottom = self.getContentsMargins()
        size += QSize(left + right, top + bottom)
        return size

    def doLayout(self, rect, testonly):
        left, top, right, bottom = self.getContentsMargins()
        effective = rect.adjusted(+left, +top, -right, -bottom)
        x = effective.x()
        y = effective.y()
        lineheight = 0
        for item in self._items:
            widget = item.widget()
            if not widget.isVisible():
                continue
            hspace = self.horizontalSpacing()
            if hspace == -1:
                hspace = widget.style().layoutSpacing(
                    QSizePolicy.PushButton, QSizePolicy.PushButton, Qt.Horizontal
                )
            vspace = self.verticalSpacing()
            if vspace == -1:
                vspace = widget.style().layoutSpacing(
                    QSizePolicy.PushButton, QSizePolicy.PushButton, Qt.Vertical
                )
            nextX = x + item.sizeHint().width() + hspace
            if nextX - hspace > effective.right() and lineheight > 0:
                x = effective.x()
                y = y + lineheight + vspace
                nextX = x + item.sizeHint().width() + hspace
                lineheight = 0
            if not testonly:
                item.setGeometry(QRect(QPoint(x, y), item.sizeHint()))
            x = nextX
            lineheight = max(lineheight, item.sizeHint().height())
        return y + lineheight - rect.y() + bottom

    def smartSpacing(self, pm):
        parent = self.parent()
        if parent is None:
            return -1
        elif parent.isWidgetType():
            return parent.style().pixelMetric(pm, None, parent)
        else:
            return parent.spacing()



import datetime
import json
import os
from logging import getLogger

from PyQt5.QtCore import QObject, QProcess, pyqtSignal, QUrl, QTimer
from PyQt5.QtGui import QDesktopServices
from PyQt5.QtNetwork import QLocalSocket
from PyQt5.QtWidgets import QMessageBox, QPushButton

from rare.components.dialogs.uninstall_dialog import UninstallDialog
from rare.components.tabs.games import CloudSaveUtils
from rare.game_launch_helper import message_models
from rare.shared import LegendaryCoreSingleton, GlobalSignalsSingleton, ArgumentsSingleton
from rare.utils import legendary_utils
from rare.utils import misc
from rare.utils.meta import RareGameMeta

logger = getLogger("GameUtils")


class GameProcess(QObject):
    game_finished = pyqtSignal(int, str)  # exit_code, appname
    game_launched = pyqtSignal(str)
    tried_connections = 0

    def __init__(self, app_name: str, on_startup=False, always_ask_sync: bool = False):
        super(GameProcess, self).__init__()
        self.app_name = app_name
        self.on_startup = on_startup
        self.game = LegendaryCoreSingleton().get_game(app_name)
        self.game_meta = RareGameMeta()
        self.socket = QLocalSocket()
        self.socket.connected.connect(self._socket_connected)
        try:
            self.socket.errorOccurred.connect(self._error_occurred)
        except AttributeError:
            QTimer.singleShot(100, lambda: self._error_occurred(None) if self.socket.error() else None)
            logger.warning("Do not handle errors on QLocalSocket, because of an old qt version")
        self.socket.readyRead.connect(self._message_available)
        self.always_ask_sync = always_ask_sync

        def close_socket():
            try:
                self.socket.close()
            except RuntimeError:
                pass

        self.socket.disconnected.connect(close_socket)
        self.timer = QTimer()
        if not on_startup:
            # wait a short time for process started
            self.timer.timeout.connect(self.connect_to_server)
            self.timer.start(200)
        else:
            # nothing happens, if no server available
            self.connect_to_server()

    def connect_to_server(self):
        self.socket.connectToServer(f"rare_{self.app_name}")
        self.tried_connections += 1

        if self.tried_connections > 50:  # 10 seconds
            QMessageBox.warning(None, "Error", self.tr("Connection to game process failed (Timeout)"))
            self.timer.stop()
            self.game_finished.emit(1, self.app_name)

    def _message_available(self):
        message = self.socket.readAll().data()
        if not message.startswith(b"{"):
            logger.error(f"Received unsupported message: {message.decode('utf-8')}")
            return
        try:
            data = json.loads(message)
        except json.JSONDecodeError as e:
            logger.error(e)
            logger.error("Could not load json data")
            return

        action = data.get("action", -1)

        if action == -1:
            logger.error("Got unexpected action")
        elif action == message_models.Actions.finished:
            logger.info(f"{self.app_name} finished")
            model = message_models.FinishedModel.from_json(data)
            self.socket.close()
            self._game_finished(model.exit_code)
        elif action == message_models.Actions.error:
            model = message_models.ErrorModel.from_json(data)
            logger.error(f"Error in game {self.game.app_title}: {model.error_string}")
            QMessageBox.warning(None, "Error", self.tr(
                "Error in game {}: \n{}").format(self.game.app_title, model.error_string))
        elif action == message_models.Actions.state_update:
            model = message_models.StateChangedModel.from_json(data)
            if model.new_state == message_models.StateChangedModel.States.started:
                logger.info("Launched Game")
                self.game_launched.emit(self.app_name)
                meta_data = self.game_meta.get_game(self.app_name)
                meta_data.last_played = datetime.datetime.now()
                self.game_meta.set_game(self.app_name, meta_data)

    def _socket_connected(self):
        self.timer.stop()
        self.timer.deleteLater()
        logger.info(f"Connection established for {self.app_name}")
        if self.on_startup:
            logger.info(f"Found {self.app_name} running at startup")

            # FIXME run this after startup, widgets do not exist at this time
            QTimer.singleShot(1000, lambda: self.game_launched.emit(self.app_name))

    def _error_occurred(self, _):
        if self.on_startup:
            self.socket.close()
            self.deleteLater()
            self._game_finished(-1234)  # 1234 is exit code for startup
        logger.error(f"{self.app_name}: {self.socket.errorString()}")

    def _game_finished(self, exit_code: int):
        self.game_finished.emit(exit_code, self.app_name)


class GameUtils(QObject):
    running_games = dict()
    finished = pyqtSignal(str, str)  # app_name, error
    cloud_save_finished = pyqtSignal(str)
    launch_queue = dict()
    game_launched = pyqtSignal(str)
    update_list = pyqtSignal(str)

    def __init__(self, parent=None):
        super(GameUtils, self).__init__(parent=parent)
        self.core = LegendaryCoreSingleton()
        self.signals = GlobalSignalsSingleton()
        self.args = ArgumentsSingleton()

        self.cloud_save_utils = CloudSaveUtils()
        self.cloud_save_utils.sync_finished.connect(self.sync_finished)
        self.game_meta = RareGameMeta()

        for igame in self.core.get_installed_list():
            game_process = GameProcess(igame.app_name, True)
            game_process.game_finished.connect(self.game_finished)
            game_process.game_launched.connect(self.game_launched.emit)
            self.running_games[igame.app_name] = game_process

    def uninstall_game(self, app_name) -> bool:
        # returns if uninstalled
        game = self.core.get_game(app_name)
        igame = self.core.get_installed_game(app_name)
        if not os.path.exists(igame.install_path):
            if QMessageBox.Yes == QMessageBox.question(
                    None,
                    self.tr("Uninstall - {}").format(igame.title),
                    self.tr(
                        "Game files of {} do not exist. Remove it from installed games?"
                    ).format(igame.title),
                    QMessageBox.Yes | QMessageBox.No,
                    QMessageBox.Yes,
            ):
                self.core.lgd.remove_installed_game(app_name)
                return True
            else:
                return False

        proceed, keep_files, keep_config = UninstallDialog(game).get_options()
        if not proceed:
            return False
        success, message = legendary_utils.uninstall_game(self.core, game.app_name, keep_files, keep_config)
        if not success:
            QMessageBox.warning(None, self.tr("Uninstall - {}").format(igame.title), message, QMessageBox.Close)
        self.signals.game_uninstalled.emit(app_name)
        return True

    def prepare_launch(
            self, app_name, offline: bool = False, skip_update_check: bool = False
    ):
        game = self.core.get_game(app_name)
        dont_sync_after_finish = False

        # TODO move this to helper
        if game.supports_cloud_saves and not offline:
            try:
                sync = self.cloud_save_utils.sync_before_launch_game(app_name)
            except ValueError:
                logger.info("Cancel startup")
                self.sync_finished(app_name)
                return
            except AssertionError:
                dont_sync_after_finish = True
            else:
                if sync:
                    self.launch_queue[app_name] = (app_name, skip_update_check, offline)
                    return
            self.sync_finished(app_name)

        self.launch_game(
            app_name, offline, skip_update_check, ask_always_sync=dont_sync_after_finish
        )

    def launch_game(
            self,
            app_name: str,
            offline: bool = False,
            skip_update_check: bool = False,
            wine_bin: str = None,
            wine_pfx: str = None,
            ask_always_sync: bool = False,
    ):
        executable = misc.get_rare_executable()
        executable, args = executable[0], executable[1:]
        args.extend([
            "start", app_name
        ])
        if offline:
            args.append("--offline")
        if skip_update_check:
            args.append("--skip-update-check")
        if wine_bin:
            args.extend(["--wine-bin", wine_bin])
        if wine_pfx:
            args.extend(["--wine-prefix", wine_pfx])
        if ask_always_sync:
            args.extend("--ask-always-sync")

        # kill me, if I don't change it before commit
        QProcess.startDetached(executable, args)
        logger.info(f"Start new Process: ({executable} {' '.join(args)})")
        game_process = GameProcess(app_name, ask_always_sync)
        game_process.game_finished.connect(self.game_finished)
        game_process.game_launched.connect(self.game_launched.emit)
        self.running_games[app_name] = game_process

    def game_finished(self, exit_code, app_name):
        if self.running_games.get(app_name):
            self.running_games.pop(app_name)
        if exit_code == -1234:
            return

        self.finished.emit(app_name, "")

        logger.info(f"Game exited with exit code: {exit_code}")
        self.signals.set_discord_rpc.emit("")
        is_origin = self.core.get_game(app_name).third_party_store == "Origin"
        if exit_code == 1 and is_origin:
            msg_box = QMessageBox()
            msg_box.setText(
                self.tr(
                    "Origin is not installed. Do you want to download installer file? "
                )
            )
            msg_box.addButton(QPushButton("Download"), QMessageBox.YesRole)
            msg_box.addButton(QPushButton("Cancel"), QMessageBox.RejectRole)
            resp = msg_box.exec()
            # click install button
            if resp == 0:
                QDesktopServices.openUrl(QUrl("https://www.dm.origin.com/download"))
            return

        if exit_code != 0:
            pass
            """
            QMessageBox.warning(
                None,
                "Warning",
                self.tr("Failed to launch {}. Check logs to find error").format(
                    self.core.get_game(app_name).app_title
                ),
            )
            """

        if app_name in self.running_games.keys():
            self.running_games.pop(app_name)

        if self.core.get_game(app_name).supports_cloud_saves:
            if exit_code != 0:
                r = QMessageBox.question(
                    None,
                    "Question",
                    self.tr(
                        "Game exited with code {}, which is not a normal code. "
                        "It could be caused by a crash. Do you want to sync cloud saves"
                    ).format(exit_code),
                    buttons=QMessageBox.Yes | QMessageBox.No,
                    defaultButton=QMessageBox.Yes,
                )
                if r != QMessageBox.Yes:
                    return

            # TODO move this to helper
            self.cloud_save_utils.game_finished(app_name, always_ask=False)

    def sync_finished(self, app_name):
        if app_name in self.launch_queue.keys():
            self.cloud_save_finished.emit(app_name)
            params = self.launch_queue[app_name]
            self.launch_queue.pop(app_name)
            self.launch_game(*params)
        else:
            self.cloud_save_finished.emit(app_name)



from logging import getLogger
from typing import Tuple, Dict, Union, List, Set

from PyQt5.QtCore import QSettings, Qt, pyqtSlot
from PyQt5.QtWidgets import QStackedWidget, QVBoxLayout, QWidget, QScrollArea, QFrame
from legendary.models.game import InstalledGame, Game

from rare.shared import (
    LegendaryCoreSingleton,
    GlobalSignalsSingleton,
    ArgumentsSingleton,
    ApiResultsSingleton,
    ImageManagerSingleton,
)
from rare.widgets.library_layout import LibraryLayout
from rare.widgets.sliding_stack import SlidingStackedWidget
from .cloud_save_utils import CloudSaveUtils
from .game_info import GameInfoTabs
from .game_info.uninstalled_info import UninstalledInfoTabs
from .game_utils import GameUtils
from .game_widgets.base_installed_widget import BaseInstalledWidget
from .game_widgets.base_uninstalled_widget import BaseUninstalledWidget
from .game_widgets.installed_icon_widget import InstalledIconWidget
from .game_widgets.installed_list_widget import InstalledListWidget
from .game_widgets.installing_game_widget import InstallingGameWidget
from .game_widgets.uninstalled_icon_widget import UninstalledIconWidget
from .game_widgets.uninstalled_list_widget import UninstalledListWidget
from .head_bar import GameListHeadBar
from .import_sync import ImportSyncTabs

logger = getLogger("GamesTab")


class GamesTab(QStackedWidget):
    def __init__(self, parent=None):
        super(GamesTab, self).__init__(parent=parent)
        self.core = LegendaryCoreSingleton()
        self.signals = GlobalSignalsSingleton()
        self.args = ArgumentsSingleton()
        self.api_results = ApiResultsSingleton()
        self.image_manager = ImageManagerSingleton()
        self.settings = QSettings()

        self.widgets: Dict[str, Tuple[
            Union[InstalledIconWidget, UninstalledIconWidget], Union[InstalledListWidget, UninstalledListWidget]]] = {}
        self.updates: Set = set()
        self.active_filter: int = 0
        self.uninstalled_games: List[Game] = []

        self.game_list: List[Game] = self.api_results.game_list
        self.dlcs = self.api_results.dlcs
        self.bit32 = self.api_results.bit32_games
        self.mac_games = self.api_results.mac_games
        self.no_assets = self.api_results.no_asset_games

        self.game_utils = GameUtils(parent=self)

        self.games = QWidget(parent=self)
        self.games.setLayout(QVBoxLayout())
        self.addWidget(self.games)

        self.head_bar = GameListHeadBar(parent=self.games)
        self.head_bar.import_clicked.connect(self.show_import)
        self.head_bar.egl_sync_clicked.connect(self.show_egl_sync)
        self.games.layout().addWidget(self.head_bar)

        self.game_info_tabs = GameInfoTabs(self.dlcs, self.game_utils, self)
        self.game_info_tabs.back_clicked.connect(lambda: self.setCurrentWidget(self.games))
        self.addWidget(self.game_info_tabs)

        self.game_info_tabs.info.verification_finished.connect(
            self.verification_finished
        )
        self.game_info_tabs.info.uninstalled.connect(lambda x: self.setCurrentWidget(self.games))

        self.import_sync_tabs = ImportSyncTabs(self)
        self.import_sync_tabs.back_clicked.connect(lambda: self.setCurrentWidget(self.games))
        self.addWidget(self.import_sync_tabs)

        for i in self.game_list:
            if i.app_name.startswith("UE_4"):
                pixmap = self.image_manager.get_pixmap(i.app_name)
                if pixmap.isNull():
                    continue
                self.ue_name = i.app_name
                logger.debug(f"Found Unreal AppName {self.ue_name}")
                break
        else:
            logger.warning("No Unreal engine in library found")
            self.ue_name = ""

        self.uninstalled_info_tabs = UninstalledInfoTabs(self)
        self.uninstalled_info_tabs.back_clicked.connect(lambda: self.setCurrentWidget(self.games))
        self.addWidget(self.uninstalled_info_tabs)

        self.no_asset_names = []
        if not self.args.offline:
            for game in self.no_assets:
                self.no_asset_names.append(game.app_name)
        else:
            self.no_assets = []

        self.installed = self.core.get_installed_list()

        self.view_stack = SlidingStackedWidget(self.games)
        self.view_stack.setFrameStyle(QFrame.NoFrame)
        self.icon_view_scroll = QScrollArea(self.view_stack)
        self.icon_view_scroll.setWidgetResizable(True)
        self.icon_view_scroll.setFrameShape(QFrame.StyledPanel)
        self.icon_view_scroll.horizontalScrollBar().setDisabled(True)
        self.list_view_scroll = QScrollArea(self.view_stack)
        self.list_view_scroll.setWidgetResizable(True)
        self.list_view_scroll.setFrameShape(QFrame.StyledPanel)
        self.list_view_scroll.horizontalScrollBar().setDisabled(True)
        self.icon_view = QWidget(self.icon_view_scroll)
        self.icon_view.setLayout(LibraryLayout(self.icon_view))
        self.icon_view.layout().setContentsMargins(0, 13, 0, 0)
        self.icon_view.layout().setAlignment(Qt.AlignTop)
        self.list_view = QWidget(self.list_view_scroll)
        self.list_view.setLayout(QVBoxLayout(self.list_view))
        self.list_view.layout().setContentsMargins(3, 3, 9, 3)
        self.list_view.layout().setAlignment(Qt.AlignTop)
        self.icon_view_scroll.setWidget(self.icon_view)
        self.list_view_scroll.setWidget(self.list_view)
        self.view_stack.addWidget(self.icon_view_scroll)
        self.view_stack.addWidget(self.list_view_scroll)
        self.games.layout().addWidget(self.view_stack)

        # add installing game widget for icon view: List view not supported
        self.installing_widget = InstallingGameWidget()
        self.icon_view.layout().addWidget(self.installing_widget)
        self.installing_widget.setVisible(False)

        if not self.settings.value("icon_view", True, bool):
            self.view_stack.setCurrentWidget(self.list_view_scroll)
            self.head_bar.view.list()
        else:
            self.view_stack.setCurrentWidget(self.icon_view_scroll)

        self.head_bar.search_bar.textChanged.connect(lambda x: self.filter_games("", x))
        self.head_bar.search_bar.textChanged.connect(
            lambda x: self.icon_view_scroll.verticalScrollBar().setSliderPosition(
                self.icon_view_scroll.verticalScrollBar().minimum()
            )
        )
        self.head_bar.search_bar.textChanged.connect(
            lambda x: self.list_view_scroll.verticalScrollBar().setSliderPosition(
                self.list_view_scroll.verticalScrollBar().minimum()
            )
        )
        self.head_bar.filterChanged.connect(self.filter_games)
        self.head_bar.refresh_list.clicked.connect(self.update_list)
        self.head_bar.view.toggled.connect(self.toggle_view)

        f = self.settings.value("filter", 0, int)
        if f >= len(self.head_bar.available_filters):
            f = 0
        self.active_filter = self.head_bar.available_filters[f]

        # signals
        self.signals.dl_progress.connect(self.installing_widget.set_status)
        self.signals.installation_started.connect(self.installation_started)
        self.signals.update_gamelist.connect(self.update_list)
        self.signals.installation_finished.connect(
            self.installation_finished
        )
        self.signals.game_uninstalled.connect(lambda name: self.update_list([name]))

        self.game_utils.update_list.connect(self.update_list)

        self.setup_game_list()

    def installation_finished(self, app_name: str):
        self.installing_widget.setVisible(False)
        self.installing_widget.set_game("")
        self.filter_games("")

    def installation_started(self, app_name: str):
        self.installing_widget.set_game(app_name)

        i_widget, l_widget = self.widgets.get(app_name, (None, None))
        if not i_widget or not l_widget:
            return
        i_widget.setVisible(False)
        l_widget.setVisible(False)

    def verification_finished(self, igame: InstalledGame):
        # only if igame needs verification
        i_widget, l_widget = self.widgets[igame.app_name]
        i_widget.igame = igame
        l_widget.igame = igame

        i_widget.leaveEvent(None)
        l_widget.update_text()

    def show_import(self):
        self.setCurrentWidget(self.import_sync_tabs)
        self.import_sync_tabs.show_import()

    def show_egl_sync(self, idx):
        self.setCurrentWidget(self.import_sync_tabs)
        self.import_sync_tabs.show_egl_sync()

    def show_game_info(self, app_name):
        self.game_info_tabs.update_game(app_name)
        self.setCurrentWidget(self.game_info_tabs)

    def show_uninstalled_info(self, game):
        self.uninstalled_info_tabs.update_game(game)
        self.setCurrentWidget(self.uninstalled_info_tabs)

    @pyqtSlot()
    def update_count_games_label(self):
        self.head_bar.set_games_count(len(self.core.get_installed_list()), len(self.game_list))

    def setup_game_list(self):
        self.update_count_games_label()

        # add installed games
        for igame in sorted(self.core.get_installed_list(), key=lambda x: x.title):
            icon_widget, list_widget = self.add_installed_widget(igame.app_name)
            self.icon_view.layout().addWidget(icon_widget)
            self.list_view.layout().addWidget(list_widget)

        for game in self.no_assets:
            if not game.metadata.get("customAttributes", {}).get("ThirdPartyManagedApp", {}).get("value") == "Origin":
                icon_widget, list_widget = self.add_uninstalled_widget(game)
            else:
                icon_widget, list_widget = self.add_installed_widget(game.app_name)
            if not icon_widget or not list_widget:
                logger.warning(f"Ignoring {game.app_name} in game list")
                continue
            self.icon_view.layout().addWidget(icon_widget)
            self.list_view.layout().addWidget(list_widget)

        for game in sorted(self.game_list, key=lambda x: x.app_title):
            if not self.core.is_installed(game.app_name):
                self.uninstalled_games.append(game)
                icon_widget, list_widget = self.add_uninstalled_widget(game)
                self.icon_view.layout().addWidget(icon_widget)
                self.list_view.layout().addWidget(list_widget)
        self.filter_games(self.active_filter)

    def add_installed_widget(self, app_name):
        pixmap = self.image_manager.get_pixmap(app_name)
        try:
            if pixmap.isNull():
                logger.info(f"{app_name} has a corrupt image.")
                if app_name in self.no_asset_names and self.core.get_asset(app_name).namespace != "ue":
                    self.image_manager.download_image_blocking(self.core.get_game(app_name), force=True)
                    pixmap = self.image_manager.get_pixmap(app_name)
                elif self.ue_name:
                    pixmap = self.image_manager.get_pixmap(self.ue_name)

            icon_widget = InstalledIconWidget(app_name, pixmap, self.game_utils)
            list_widget = InstalledListWidget(app_name, pixmap, self.game_utils)
        except Exception as e:
            logger.error(f"{app_name} is broken. Don't add it to game list: {e}")
            return None, None

        self.widgets[app_name] = (icon_widget, list_widget)

        icon_widget.show_info.connect(self.show_game_info)
        list_widget.show_info.connect(self.show_game_info)

        if icon_widget.update_available:
            self.updates.add(app_name)
        return icon_widget, list_widget

    def add_uninstalled_widget(self, game):
        pixmap = self.image_manager.get_pixmap(game.app_name, color=False)
        try:
            if pixmap.isNull():
                if self.core.get_asset(game.app_name).namespace != "ue":
                    logger.warning(f"{game.app_title} has a corrupt image. Reloading...")
                    self.image_manager.download_image_blocking(game, force=True)
                    pixmap = self.image_manager.get_pixmap(game.app_name, color=False)
                elif self.ue_name:
                    pixmap = self.image_manager.get_pixmap(self.ue_name, color=False)

            icon_widget = UninstalledIconWidget(game, self.core, pixmap)
            list_widget = UninstalledListWidget(self.core, game, pixmap)
        except Exception as e:
            logger.error(f"{game.app_name} is broken. Don't add it to game list: {e}")
            return None, None

        icon_widget.show_uninstalled_info.connect(self.show_uninstalled_info)
        list_widget.show_uninstalled_info.connect(self.show_uninstalled_info)

        self.widgets[game.app_name] = (icon_widget, list_widget)

        return icon_widget, list_widget

    def filter_games(self, filter_name="all", search_text: str = ""):
        if not search_text and (t := self.head_bar.search_bar.text()):
            search_text = t

        if filter_name:
            self.active_filter = filter_name
        if not filter_name and (t := self.active_filter):
            filter_name = t

        def get_visibility(widget) -> Tuple[bool, float]:
            app_name = widget.game.app_name

            if not isinstance(widget,
                              InstallingGameWidget) and self.installing_widget.game and self.installing_widget.game.app_name == app_name:
                visible = False
            elif filter_name == "installed":
                visible = self.core.is_installed(app_name)
            elif filter_name == "offline":
                if self.core.is_installed(app_name) and not isinstance(widget, InstallingGameWidget):
                    visible = widget.igame.can_run_offline
                else:
                    visible = False
            elif filter_name == "32bit" and self.bit32:
                visible = app_name in self.bit32
            elif filter_name == "mac" and self.mac_games:
                visible = app_name in self.mac_games
            elif filter_name == "installable":
                visible = app_name not in self.no_asset_names
            elif filter_name == "include_ue":
                visible = True
            elif filter_name == "all":
                # All visible except ue
                try:
                    visible = self.core.get_asset(app_name, update=False).namespace != "ue"
                except ValueError:
                    visible = True
            else:
                visible = True

            if (
                    search_text.lower() not in widget.game.app_name.lower()
                    and search_text.lower() not in widget.game.app_title.lower()
            ):
                opacity = 0.25
            else:
                opacity = 1.0
            return visible, opacity

        for t in self.widgets.values():
            visible, opacity = get_visibility(t[0])
            for w in t:
                w.setVisible(visible)
                w.image.setOpacity(opacity)

        self.sort_list(search_text)

        if self.installing_widget.game:
            self.installing_widget.setVisible(get_visibility(self.installing_widget)[0])

    @pyqtSlot()
    def sort_list(self, sort_by: str = ""):
        # lk: this is the existing sorting implemenation
        # lk: it sorts by installed then by title
        installing_widget = self.icon_view.layout().remove(type(self.installing_widget).__name__)
        if sort_by:
            self.icon_view.layout().sort(lambda x: (sort_by.lower() not in x.widget().game.app_title.lower(),))
        else:
            self.icon_view.layout().sort(
                lambda x: (
                    not x.widget().is_installed,
                    x.widget().is_non_asset,
                    x.widget().app_title,
                )
            )
        self.icon_view.layout().insert(0, installing_widget)
        list_widgets = self.list_view.findChildren(InstalledListWidget) + self.list_view.findChildren(
            UninstalledListWidget)
        if sort_by:
            list_widgets.sort(key=lambda x: (sort_by not in x.game.app_title.lower(),))
        else:
            list_widgets.sort(
                key=lambda x: (not x.is_installed, x.is_non_asset, x.app_title)
            )
        for idx, wl in enumerate(list_widgets):
            self.list_view.layout().insertWidget(idx, wl)

    def __remove_from_layout(self, app_name):
        self.icon_view.layout().removeWidget(self.widgets[app_name][0])
        self.list_view.layout().removeWidget(self.widgets[app_name][1])
        self.widgets[app_name][0].deleteLater()
        self.widgets[app_name][1].deleteLater()
        self.widgets.pop(app_name)

    def __update_installed(self, app_name):
        self.__remove_from_layout(app_name)
        icon_widget, list_widget = self.add_installed_widget(app_name)
        self.icon_view.layout().addWidget(icon_widget)
        self.list_view.layout().addWidget(list_widget)

    def __update_uninstalled(self, app_name):
        self.__remove_from_layout(app_name)
        game = self.core.get_game(app_name, False)
        try:
            icon_widget, list_widget = self.add_uninstalled_widget(game)
            self.icon_view.layout().addWidget(icon_widget)
            self.list_view.layout().addWidget(list_widget)
        except Exception:
            pass

    def update_list(self, app_names: list = None):
        logger.debug(f"Updating list for {app_names}")
        if app_names:
            update_list = False
            for app_name in app_names:
                if widgets := self.widgets.get(app_name):

                    # from update
                    if self.core.is_installed(widgets[0].game.app_name) and isinstance(
                            widgets[0], BaseInstalledWidget
                    ):
                        logger.debug(f"Update Gamelist: Updated: {app_name}")
                        igame = self.core.get_installed_game(app_name)
                        for w in widgets:
                            w.igame = igame
                            w.update_available = (
                                    self.core.get_asset(
                                        w.game.app_name, w.igame.platform, False
                                    ).build_version
                                    != igame.version
                            )
                        widgets[0].leaveEvent(None)
                    # new installed
                    elif self.core.is_installed(app_name) and isinstance(
                            widgets[0], BaseUninstalledWidget
                    ):
                        logger.debug(f"Update Gamelist: New installed {app_name}")
                        self.__update_installed(app_name)
                        update_list = True

                    # uninstalled
                    elif not self.core.is_installed(
                            widgets[0].game.app_name
                    ) and isinstance(widgets[0], BaseInstalledWidget):
                        logger.debug(f"Update list: Uninstalled: {app_name}")
                        self.__update_uninstalled(app_name)
                        update_list = True

            # do not update, if only update finished
            if update_list:
                self.sort_list()

        else:
            installed_names = [i.app_name for i in self.core.get_installed_list()]
            # get Uninstalled games
            uninstalled_names = []
            games = self.core.get_game_list(True)
            for game in sorted(games, key=lambda x: x.app_title):
                if not game.app_name in installed_names:
                    uninstalled_names.append(game.app_name)

            new_installed_games = list(
                set(installed_names) - set([i.app_name for i in self.installed])
            )
            new_uninstalled_games = list(
                set(uninstalled_names)
                - set([i.app_name for i in self.uninstalled_games])
            )

            if (not new_uninstalled_games) and (not new_installed_games):
                return

            if new_installed_games:
                for name in new_installed_games:
                    self.__update_installed(name)

                for name in new_uninstalled_games:
                    self.__update_uninstalled(name)

                self.sort_list()
        self.update_count_games_label()

    def toggle_view(self):
        self.settings.setValue("icon_view", not self.head_bar.view.isChecked())

        if not self.head_bar.view.isChecked():
            self.view_stack.slideInWidget(self.icon_view_scroll)
        else:
            self.view_stack.slideInWidget(self.list_view_scroll)



from PyQt5.QtCore import QParallelAnimationGroup, Qt, QPropertyAnimation, QAbstractAnimation
from PyQt5.QtWidgets import QWidget, QFrame, QToolButton, QGridLayout, QSizePolicy, QLayout

from rare.utils.misc import icon

# https://newbedev.com/how-to-make-an-expandable-collapsable-section-widget-in-qt


class CollabsibleWidget(QWidget):
    def __init__(
        self, child_layout: QLayout = None, title: str = "", animation_duration: int = 200, parent=None
    ):
        """
        References:
            # Adapted from c++ version
            https://stackoverflow.com/questions/32476006/how-to-make-an-expandable-collapsable-section-widget-in-qt
        """
        super(CollabsibleWidget, self).__init__(parent=parent)

        self.animationDuration = animation_duration
        self.toggleAnimation = QParallelAnimationGroup()
        self.contentArea = QWidget()
        self.headerLine = QFrame()
        self.toggleButton = QToolButton()
        self.mainLayout = QGridLayout()

        toggleButton = self.toggleButton
        toggleButton.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
        toggleButton.setIcon(icon("fa.arrow-right"))
        toggleButton.setText(str(title))
        toggleButton.setCheckable(True)
        toggleButton.setChecked(False)

        headerLine = self.headerLine
        headerLine.setFrameShape(QFrame.StyledPanel)
        headerLine.setFrameShadow(QFrame.Plain)
        headerLine.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Maximum)

        self.contentArea.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        # start out collapsed
        self.contentArea.setMaximumHeight(0)
        self.contentArea.setMinimumHeight(0)
        # let the entire widget grow and shrink with its content
        toggleAnimation = self.toggleAnimation
        toggleAnimation.addAnimation(QPropertyAnimation(self, b"minimumHeight"))
        toggleAnimation.addAnimation(QPropertyAnimation(self, b"maximumHeight"))
        toggleAnimation.addAnimation(QPropertyAnimation(self.contentArea, b"maximumHeight"))
        # don't waste space
        mainLayout = self.mainLayout
        mainLayout.setVerticalSpacing(0)
        mainLayout.setContentsMargins(0, 0, 0, 0)
        row = 0
        mainLayout.addWidget(self.toggleButton, row, 0, 1, 1, Qt.AlignLeft)
        mainLayout.addWidget(self.headerLine, row, 2, 1, 1)
        row += 1
        mainLayout.addWidget(self.contentArea, row, 0, 1, 3)
        self.setLayout(self.mainLayout)

        def start_animation(checked):
            arrow_type = icon("fa.arrow-down") if checked else icon("fa.arrow-right")
            direction = QAbstractAnimation.Forward if checked else QAbstractAnimation.Backward
            toggleButton.setIcon(arrow_type)
            self.toggleAnimation.setDirection(direction)
            self.toggleAnimation.start()

        self.toggleButton.clicked.connect(start_animation)

        if child_layout:
            self.setContentLayout(child_layout)

    def setContentLayout(self, content_layout: QLayout):
        # Not sure if this is equivalent to self.contentArea.destroy()
        self.contentArea.destroy()
        self.contentArea.setLayout(content_layout)
        collapsedHeight = self.sizeHint().height() - self.contentArea.maximumHeight()
        contentHeight = content_layout.sizeHint().height()
        for i in range(self.toggleAnimation.animationCount() - 1):
            spoilerAnimation = self.toggleAnimation.animationAt(i)
            spoilerAnimation.setDuration(self.animationDuration)
            spoilerAnimation.setStartValue(collapsedHeight)
            spoilerAnimation.setEndValue(collapsedHeight + contentHeight)
        contentAnimation = self.toggleAnimation.animationAt(self.toggleAnimation.animationCount() - 1)
        contentAnimation.setDuration(self.animationDuration)
        contentAnimation.setStartValue(0)
        contentAnimation.setEndValue(contentHeight)



import platform

from PyQt5.QtCore import Qt, QThreadPool
from PyQt5.QtGui import QKeyEvent
from PyQt5.QtWidgets import QWidget, QTreeView
from legendary.models.game import Game

from rare.shared import (
    LegendaryCoreSingleton,
    GlobalSignalsSingleton,
    ArgumentsSingleton,
    ApiResultsSingleton,
)
from rare.shared import ImageManagerSingleton
from rare.shared.image_manager import ImageSize
from rare.ui.components.tabs.games.game_info.game_info import Ui_GameInfo
from rare.utils.extra_widgets import SideTabWidget
from rare.utils.json_formatter import QJsonModel
from rare.models.install import InstallOptionsModel
from rare.utils.steam_grades import SteamWorker
from rare.widgets.image_widget import ImageWidget


class UninstalledInfoTabs(SideTabWidget):
    def __init__(self, parent=None):
        super(UninstalledInfoTabs, self).__init__(show_back=True, parent=parent)
        self.core = LegendaryCoreSingleton()
        self.signals = GlobalSignalsSingleton()
        self.args = ArgumentsSingleton()

        self.info = UninstalledInfo()
        self.info.install_button.setDisabled(self.args.offline)
        self.addTab(self.info, self.tr("Information"))

        self.view = GameMetadataView()
        self.addTab(self.view, self.tr("Metadata"))

        # self.setTabEnabled(1, False)
        self.setCurrentIndex(1)

    def update_game(self, game: Game):
        self.setCurrentIndex(1)
        self.info.update_game(game)
        self.view.update_game(game)

    def keyPressEvent(self, e: QKeyEvent):
        if e.key() == Qt.Key_Escape:
            self.parent().setCurrentIndex(0)


class GameMetadataView(QTreeView):
    game: Game

    def __init__(self, parent=None):
        super(GameMetadataView, self).__init__(parent=parent)
        self.setColumnWidth(0, 300)
        self.setWordWrap(True)
        self.model = QJsonModel()
        self.setModel(self.model)

    def update_game(self, game: Game):
        self.game = game
        self.title.setTitle(self.game.app_title)
        self.model.clear()
        try:
            self.model.load(game.__dict__)
        except:
            pass
        self.resizeColumnToContents(0)


class UninstalledInfo(QWidget, Ui_GameInfo):
    game: Game

    def __init__(self, parent=None):
        super(UninstalledInfo, self).__init__(parent=parent)
        self.setupUi(self)
        self.core = LegendaryCoreSingleton()
        self.signals = GlobalSignalsSingleton()
        self.api_results = ApiResultsSingleton()
        self.image_manager = ImageManagerSingleton()

        self.image = ImageWidget(self)
        self.image.setFixedSize(ImageSize.Display)
        self.layout_game_info.insertWidget(0, self.image, alignment=Qt.AlignTop)

        self.install_button.clicked.connect(self.install_game)
        if platform.system() != "Windows":
            self.steam_worker = SteamWorker(self.core)
            self.steam_worker.signals.rating.connect(self.grade.setText)
            self.steam_worker.setAutoDelete(False)

        else:
            self.lbl_grade.setVisible(False)
            self.grade.setVisible(False)

        self.install_size.setEnabled(False)
        self.lbl_install_size.setEnabled(False)
        self.install_path.setEnabled(False)
        self.lbl_install_path.setEnabled(False)

        self.game_actions_stack.setCurrentIndex(1)
        self.game_actions_stack.resize(self.game_actions_stack.minimumSize())
        self.lbl_platform.setText(self.tr("Platforms"))

    def install_game(self):
        self.signals.install_game.emit(InstallOptionsModel(app_name=self.game.app_name))

    def update_game(self, game: Game):
        self.game = game
        self.title.setTitle(self.game.app_title)
        available_platforms = ["Windows"]
        if self.game.app_name in self.api_results.bit32_games:
            available_platforms.append("32 Bit")
        if self.game.app_name in self.api_results.mac_games:
            available_platforms.append("macOS")
        self.platform.setText(", ".join(available_platforms))

        self.image.setPixmap(self.image_manager.get_pixmap(self.game.app_name, color=True))

        self.app_name.setText(self.game.app_name)
        self.version.setText(self.game.app_version("Windows"))
        self.dev.setText(self.game.metadata["developer"])
        self.install_size.setText("N/A")
        self.install_path.setText("N/A")

        is_ue = self.core.get_asset(game.app_name).namespace == "ue"
        self.grade.setVisible(not is_ue)
        self.lbl_grade.setVisible(not is_ue)

        if platform.system() != "Windows" and not is_ue:
            self.grade.setText(self.tr("Loading"))
            self.steam_worker.set_app_name(game.app_name)
            QThreadPool.globalInstance().start(self.steam_worker)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/dialogs/login/browser_login.ui'
#
# Created by: PyQt5 UI code generator 5.15.7
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_BrowserLogin(object):
    def setupUi(self, BrowserLogin):
        BrowserLogin.setObjectName("BrowserLogin")
        BrowserLogin.resize(182, 210)
        BrowserLogin.setWindowTitle("BrowserLogin")
        self.browser_layout = QtWidgets.QGridLayout(BrowserLogin)
        self.browser_layout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.browser_layout.setObjectName("browser_layout")
        self.link_layout = QtWidgets.QHBoxLayout()
        self.link_layout.setObjectName("link_layout")
        self.link_text = QtWidgets.QLineEdit(BrowserLogin)
        self.link_text.setText("")
        self.link_text.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop)
        self.link_text.setReadOnly(True)
        self.link_text.setPlaceholderText("")
        self.link_text.setObjectName("link_text")
        self.link_layout.addWidget(self.link_text)
        self.copy_button = QtWidgets.QPushButton(BrowserLogin)
        self.copy_button.setText("")
        self.copy_button.setObjectName("copy_button")
        self.link_layout.addWidget(self.copy_button)
        self.browser_layout.addLayout(self.link_layout, 2, 0, 1, 2)
        spacerItem = QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.browser_layout.addItem(spacerItem, 4, 0, 1, 2)
        self.open_button = QtWidgets.QPushButton(BrowserLogin)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.open_button.sizePolicy().hasHeightForWidth())
        self.open_button.setSizePolicy(sizePolicy)
        self.open_button.setObjectName("open_button")
        self.browser_layout.addWidget(self.open_button, 1, 0, 1, 1)
        self.sid_layout = QtWidgets.QHBoxLayout()
        self.sid_layout.setObjectName("sid_layout")
        self.browser_layout.addLayout(self.sid_layout, 1, 1, 1, 1)
        self.title_label = QtWidgets.QLabel(BrowserLogin)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.title_label.setFont(font)
        self.title_label.setObjectName("title_label")
        self.browser_layout.addWidget(self.title_label, 0, 0, 1, 2)
        self.info_label = QtWidgets.QLabel(BrowserLogin)
        font = QtGui.QFont()
        font.setItalic(True)
        self.info_label.setFont(font)
        self.info_label.setWordWrap(True)
        self.info_label.setObjectName("info_label")
        self.browser_layout.addWidget(self.info_label, 5, 0, 1, 2)
        self.status_label = QtWidgets.QLabel(BrowserLogin)
        self.status_label.setText("")
        self.status_label.setAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft)
        self.status_label.setObjectName("status_label")
        self.browser_layout.addWidget(self.status_label, 3, 0, 1, 2)

        self.retranslateUi(BrowserLogin)
        QtCore.QMetaObject.connectSlotsByName(BrowserLogin)

    def retranslateUi(self, BrowserLogin):
        _translate = QtCore.QCoreApplication.translate
        self.open_button.setText(_translate("BrowserLogin", "Open Browser"))
        self.title_label.setText(_translate("BrowserLogin", "Login through browser"))
        self.info_label.setText(_translate("BrowserLogin", "Click the button to open the login page in a browser or copy the link and paste it in a browser. After logging in, copy the <b><code>authorizationCode</code></b> in the input above."))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    BrowserLogin = QtWidgets.QWidget()
    ui = Ui_BrowserLogin()
    ui.setupUi(BrowserLogin)
    BrowserLogin.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/tabs/settings/widgets/eos_widget.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_EosWidget(object):
    def setupUi(self, EosWidget):
        EosWidget.setObjectName("EosWidget")
        EosWidget.resize(364, 218)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(EosWidget.sizePolicy().hasHeightForWidth())
        EosWidget.setSizePolicy(sizePolicy)
        EosWidget.setWindowTitle("GroupBox")
        self.eos_layout = QtWidgets.QHBoxLayout(EosWidget)
        self.eos_layout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.eos_layout.setObjectName("eos_layout")
        self.info_stack = QtWidgets.QStackedWidget(EosWidget)
        self.info_stack.setObjectName("info_stack")
        self.installed_info_page = QtWidgets.QWidget()
        self.installed_info_page.setObjectName("installed_info_page")
        self.installed_info_page_layout = QtWidgets.QVBoxLayout(self.installed_info_page)
        self.installed_info_page_layout.setContentsMargins(0, 0, 0, 0)
        self.installed_info_page_layout.setObjectName("installed_info_page_layout")
        self.installed_info_gb = QtWidgets.QGroupBox(self.installed_info_page)
        self.installed_info_gb.setObjectName("installed_info_gb")
        self.installed_info_layout = QtWidgets.QFormLayout(self.installed_info_gb)
        self.installed_info_layout.setLabelAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.installed_info_layout.setObjectName("installed_info_layout")
        self.installed_version_info_lbl = QtWidgets.QLabel(self.installed_info_gb)
        self.installed_version_info_lbl.setObjectName("installed_version_info_lbl")
        self.installed_info_layout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.installed_version_info_lbl)
        self.installed_version_lbl = QtWidgets.QLabel(self.installed_info_gb)
        self.installed_version_lbl.setText("TextLabel")
        self.installed_version_lbl.setObjectName("installed_version_lbl")
        self.installed_info_layout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.installed_version_lbl)
        self.installed_path_info_lbl = QtWidgets.QLabel(self.installed_info_gb)
        self.installed_path_info_lbl.setObjectName("installed_path_info_lbl")
        self.installed_info_layout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.installed_path_info_lbl)
        self.installed_path_lbl = QtWidgets.QLabel(self.installed_info_gb)
        self.installed_path_lbl.setText("TextLabel")
        self.installed_path_lbl.setObjectName("installed_path_lbl")
        self.installed_info_layout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.installed_path_lbl)
        self.update_available_info_label = QtWidgets.QLabel(self.installed_info_gb)
        self.update_available_info_label.setObjectName("update_available_info_label")
        self.installed_info_layout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.update_available_info_label)
        self.update_check_button = QtWidgets.QPushButton(self.installed_info_gb)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.update_check_button.sizePolicy().hasHeightForWidth())
        self.update_check_button.setSizePolicy(sizePolicy)
        self.update_check_button.setMaximumSize(QtCore.QSize(150, 16777215))
        self.update_check_button.setObjectName("update_check_button")
        self.installed_info_layout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.update_check_button)
        self.uninstall_info_label = QtWidgets.QLabel(self.installed_info_gb)
        self.uninstall_info_label.setObjectName("uninstall_info_label")
        self.installed_info_layout.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.uninstall_info_label)
        self.uninstall_button = QtWidgets.QPushButton(self.installed_info_gb)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.uninstall_button.sizePolicy().hasHeightForWidth())
        self.uninstall_button.setSizePolicy(sizePolicy)
        self.uninstall_button.setMaximumSize(QtCore.QSize(150, 16777215))
        self.uninstall_button.setObjectName("uninstall_button")
        self.installed_info_layout.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.uninstall_button)
        self.update_button = QtWidgets.QPushButton(self.installed_info_gb)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.update_button.sizePolicy().hasHeightForWidth())
        self.update_button.setSizePolicy(sizePolicy)
        self.update_button.setMaximumSize(QtCore.QSize(150, 16777215))
        self.update_button.setObjectName("update_button")
        self.installed_info_layout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.update_button)
        self.update_info_lbl = QtWidgets.QLabel(self.installed_info_gb)
        self.update_info_lbl.setObjectName("update_info_lbl")
        self.installed_info_layout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.update_info_lbl)
        self.installed_info_page_layout.addWidget(self.installed_info_gb, 0, QtCore.Qt.AlignTop)
        self.info_stack.addWidget(self.installed_info_page)
        self.install_overlay_page = QtWidgets.QWidget()
        self.install_overlay_page.setObjectName("install_overlay_page")
        self.install_overlay_page_layout = QtWidgets.QVBoxLayout(self.install_overlay_page)
        self.install_overlay_page_layout.setContentsMargins(0, 0, 0, 0)
        self.install_overlay_page_layout.setObjectName("install_overlay_page_layout")
        self.install_overlay_gb = QtWidgets.QGroupBox(self.install_overlay_page)
        self.install_overlay_gb.setObjectName("install_overlay_gb")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.install_overlay_gb)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.label = QtWidgets.QLabel(self.install_overlay_gb)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setObjectName("label")
        self.verticalLayout_4.addWidget(self.label)
        self.install_button = QtWidgets.QPushButton(self.install_overlay_gb)
        self.install_button.setMinimumSize(QtCore.QSize(120, 0))
        self.install_button.setObjectName("install_button")
        self.verticalLayout_4.addWidget(self.install_button, 0, QtCore.Qt.AlignRight)
        self.install_overlay_page_layout.addWidget(self.install_overlay_gb, 0, QtCore.Qt.AlignTop)
        self.info_stack.addWidget(self.install_overlay_page)
        self.eos_layout.addWidget(self.info_stack)
        self.enable_gb = QtWidgets.QGroupBox(EosWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.enable_gb.sizePolicy().hasHeightForWidth())
        self.enable_gb.setSizePolicy(sizePolicy)
        self.enable_gb.setObjectName("enable_gb")
        self.enable_layout = QtWidgets.QVBoxLayout(self.enable_gb)
        self.enable_layout.setObjectName("enable_layout")
        self.select_pfx_combo = QtWidgets.QComboBox(self.enable_gb)
        self.select_pfx_combo.setObjectName("select_pfx_combo")
        self.enable_layout.addWidget(self.select_pfx_combo)
        self.enabled_cb = QtWidgets.QCheckBox(self.enable_gb)
        self.enabled_cb.setObjectName("enabled_cb")
        self.enable_layout.addWidget(self.enabled_cb)
        self.enabled_info_label = QtWidgets.QLabel(self.enable_gb)
        font = QtGui.QFont()
        font.setItalic(True)
        self.enabled_info_label.setFont(font)
        self.enabled_info_label.setText("")
        self.enabled_info_label.setObjectName("enabled_info_label")
        self.enable_layout.addWidget(self.enabled_info_label, 0, QtCore.Qt.AlignTop)
        self.enable_layout.setStretch(2, 1)
        self.eos_layout.addWidget(self.enable_gb)

        self.retranslateUi(EosWidget)
        self.info_stack.setCurrentIndex(1)
        QtCore.QMetaObject.connectSlotsByName(EosWidget)

    def retranslateUi(self, EosWidget):
        _translate = QtCore.QCoreApplication.translate
        EosWidget.setTitle(_translate("EosWidget", "Epic Overlay settings"))
        self.installed_info_gb.setTitle(_translate("EosWidget", "Installed Info"))
        self.installed_version_info_lbl.setText(_translate("EosWidget", "Installed version"))
        self.installed_path_info_lbl.setText(_translate("EosWidget", "Installed path"))
        self.update_available_info_label.setText(_translate("EosWidget", "Updates"))
        self.update_check_button.setText(_translate("EosWidget", "Check for Update"))
        self.uninstall_info_label.setText(_translate("EosWidget", "Uninstall"))
        self.uninstall_button.setText(_translate("EosWidget", "Uninstall"))
        self.update_button.setText(_translate("EosWidget", "Update"))
        self.update_info_lbl.setText(_translate("EosWidget", "Install Update"))
        self.install_overlay_gb.setTitle(_translate("EosWidget", "Install Overlay"))
        self.label.setText(_translate("EosWidget", "No overlays are installed"))
        self.install_button.setText(_translate("EosWidget", "Install"))
        self.enable_gb.setTitle(_translate("EosWidget", "Enable / Disable"))
        self.enabled_cb.setText(_translate("EosWidget", "Activated"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    EosWidget = QtWidgets.QGroupBox()
    ui = Ui_EosWidget()
    ui.setupUi(EosWidget)
    EosWidget.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/tabs/store/browse_games.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtWidgets


class Ui_browse_games(object):
    def setupUi(self, browse_games):
        browse_games.setObjectName("browse_games")
        browse_games.resize(706, 541)
        browse_games.setWindowTitle("Form")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(browse_games)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.stack = QtWidgets.QStackedWidget(browse_games)
        self.stack.setObjectName("stack")
        self.games_page = QtWidgets.QWidget()
        self.games_page.setObjectName("games_page")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.games_page)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.games = QtWidgets.QScrollArea(self.games_page)
        self.games.setWidgetResizable(True)
        self.games.setObjectName("games")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 462, 503))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.games.setWidget(self.scrollAreaWidgetContents)
        self.verticalLayout_5.addWidget(self.games)
        self.stack.addWidget(self.games_page)
        self.error = QtWidgets.QWidget()
        self.error.setObjectName("error")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.error)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.error_label = QtWidgets.QLabel(self.error)
        self.error_label.setObjectName("error_label")
        self.verticalLayout_6.addWidget(self.error_label)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_6.addItem(spacerItem)
        self.stack.addWidget(self.error)
        self.horizontalLayout_2.addWidget(self.stack)
        self.filter_scroll = QtWidgets.QScrollArea(browse_games)
        self.filter_scroll.setMaximumSize(QtCore.QSize(200, 16777215))
        self.filter_scroll.setWidgetResizable(True)
        self.filter_scroll.setObjectName("filter_scroll")
        self.scroll_widget = QtWidgets.QWidget()
        self.scroll_widget.setGeometry(QtCore.QRect(0, 0, 198, 521))
        self.scroll_widget.setObjectName("scroll_widget")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.scroll_widget)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.price_gb = QtWidgets.QGroupBox(self.scroll_widget)
        self.price_gb.setObjectName("price_gb")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.price_gb)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.clear_price = QtWidgets.QRadioButton(self.price_gb)
        self.clear_price.setChecked(True)
        self.clear_price.setObjectName("clear_price")
        self.verticalLayout_2.addWidget(self.clear_price)
        self.free_button = QtWidgets.QRadioButton(self.price_gb)
        self.free_button.setObjectName("free_button")
        self.verticalLayout_2.addWidget(self.free_button)
        self.under10 = QtWidgets.QRadioButton(self.price_gb)
        self.under10.setObjectName("under10")
        self.verticalLayout_2.addWidget(self.under10)
        self.under20 = QtWidgets.QRadioButton(self.price_gb)
        self.under20.setObjectName("under20")
        self.verticalLayout_2.addWidget(self.under20)
        self.under30 = QtWidgets.QRadioButton(self.price_gb)
        self.under30.setObjectName("under30")
        self.verticalLayout_2.addWidget(self.under30)
        self.above = QtWidgets.QRadioButton(self.price_gb)
        self.above.setObjectName("above")
        self.verticalLayout_2.addWidget(self.above)
        self.on_discount = QtWidgets.QRadioButton(self.price_gb)
        self.on_discount.setObjectName("on_discount")
        self.verticalLayout_2.addWidget(self.on_discount)
        self.verticalLayout_3.addWidget(self.price_gb)
        self.genre_gb = QtWidgets.QGroupBox(self.scroll_widget)
        self.genre_gb.setObjectName("genre_gb")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.genre_gb)
        self.verticalLayout.setObjectName("verticalLayout")
        self.verticalLayout_3.addWidget(self.genre_gb)
        self.type_gb = QtWidgets.QGroupBox(self.scroll_widget)
        self.type_gb.setObjectName("type_gb")
        self.verticalLayout_8 = QtWidgets.QVBoxLayout(self.type_gb)
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.verticalLayout_3.addWidget(self.type_gb)
        self.platform_gb = QtWidgets.QGroupBox(self.scroll_widget)
        self.platform_gb.setObjectName("platform_gb")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.platform_gb)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.verticalLayout_3.addWidget(self.platform_gb)
        self.others_gb = QtWidgets.QGroupBox(self.scroll_widget)
        self.others_gb.setObjectName("others_gb")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.others_gb)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.verticalLayout_3.addWidget(self.others_gb)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_3.addItem(spacerItem1)
        self.filter_scroll.setWidget(self.scroll_widget)
        self.horizontalLayout_2.addWidget(self.filter_scroll)

        self.retranslateUi(browse_games)
        self.stack.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(browse_games)

    def retranslateUi(self, browse_games):
        _translate = QtCore.QCoreApplication.translate
        self.error_label.setText(_translate("browse_games", "An error occured"))
        self.price_gb.setTitle(_translate("browse_games", "Price"))
        self.clear_price.setText(_translate("browse_games", "Clear price filter"))
        self.free_button.setText(_translate("browse_games", "Free"))
        self.under10.setText(_translate("browse_games", "Under 10"))
        self.under20.setText(_translate("browse_games", "Under 20"))
        self.under30.setText(_translate("browse_games", "Under 30"))
        self.above.setText(_translate("browse_games", "14.99 and above"))
        self.on_discount.setText(_translate("browse_games", "Discount"))
        self.genre_gb.setTitle(_translate("browse_games", "Genre"))
        self.type_gb.setTitle(_translate("browse_games", "Type"))
        self.platform_gb.setTitle(_translate("browse_games", "Platform"))
        self.others_gb.setTitle(_translate("browse_games", "Other Tags"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    browse_games = QtWidgets.QWidget()
    ui = Ui_browse_games()
    ui.setupUi(browse_games)
    browse_games.show()
    sys.exit(app.exec_())



from logging import getLogger

from PyQt5.QtCore import QEvent, pyqtSignal, Qt
from PyQt5.QtGui import QMouseEvent
from PyQt5.QtWidgets import QVBoxLayout, QHBoxLayout, QPushButton, QWidget

from rare.components.tabs.games.game_widgets.base_installed_widget import (
    BaseInstalledWidget,
)
from rare.shared import LegendaryCoreSingleton
from rare.shared.image_manager import ImageSize
from rare.utils.misc import icon
from rare.widgets.elide_label import ElideLabel

logger = getLogger("GameWidgetInstalled")


class InstalledIconWidget(BaseInstalledWidget):
    update_game = pyqtSignal()

    def __init__(self, app_name, pixmap, game_utils):
        super(InstalledIconWidget, self).__init__(app_name, pixmap, game_utils)
        self.setObjectName(type(self).__name__)

        self.setContextMenuPolicy(Qt.ActionsContextMenu)
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setFixedWidth(ImageSize.Display.size.width())
        self.core = LegendaryCoreSingleton()

        if self.update_available:
            logger.info(f"Update available for game: {self.game.app_name}")

        layout.addWidget(self.image)

        self.game_utils.finished.connect(self.game_finished)
        self.game_utils.cloud_save_finished.connect(self.sync_finished)

        miniwidget = QWidget(self)
        miniwidget.setFixedWidth(ImageSize.Display.size.width())
        minilayout = QHBoxLayout()
        minilayout.setContentsMargins(0, 0, 0, 0)
        minilayout.setSpacing(0)
        miniwidget.setLayout(minilayout)

        self.title_label = ElideLabel(f"<b>{self.game.app_title}</b>", parent=miniwidget)
        self.title_label.setAlignment(Qt.AlignTop)
        self.title_label.setObjectName("game_widget")
        minilayout.addWidget(self.title_label, stretch=2)

        # Info Button
        self.menu_btn = QPushButton(parent=miniwidget)
        self.menu_btn.setIcon(icon("ei.info-circle"))
        # self.menu_btn.setObjectName("installed_menu_button")
        self.menu_btn.enterEvent = lambda x: self.info_label.setText(self.tr("Information"))
        self.menu_btn.leaveEvent = lambda x: self.enterEvent(None)
        # remove Border

        self.menu_btn.setObjectName("menu_button")

        self.menu_btn.clicked.connect(lambda: self.show_info.emit(self.game.app_name))
        self.menu_btn.setFixedSize(22, 22)
        minilayout.addWidget(self.menu_btn, stretch=0)
        minilayout.setAlignment(Qt.AlignTop)
        layout.addWidget(miniwidget)

        self.info_label = ElideLabel(" ", parent=self)
        self.info_label.setFixedWidth(ImageSize.Display.size.width())
        self.leaveEvent(None)
        self.info_label.setObjectName("info_label")
        layout.addWidget(self.info_label)

        if self.igame and self.igame.needs_verification:
            self.info_label.setText(self.texts["needs_verification"])

        self.setLayout(layout)

        self.game_utils.game_launched.connect(self.game_started)

    def enterEvent(self, a0: QEvent = None) -> None:
        if self.game_running:
            self.info_label.setText(self.texts["hover"]["running"])
        elif self.igame and self.igame.needs_verification:
            self.info_label.setText(self.texts["needs_verification"])
        elif self.is_only_offline:
            self.info_label.setText(self.texts["hover"]["launch_offline"])
        elif self.update_available:
            self.info_label.setText(self.texts["hover"]["update_available"])
        else:
            self.info_label.setText(
                self.texts["hover"]["launch" if self.igame else "launch_origin" if self.is_origin else "no_launch"]
            )

    def leaveEvent(self, a0: QEvent = None) -> None:
        if self.game_running:
            self.info_label.setText(self.texts["default"]["running"])
        elif self.syncing_cloud_saves:
            self.info_label.setText(self.texts["default"]["syncing"])
        elif self.is_only_offline:
            self.info_label.setText(self.texts["default"]["no_meta"])
        elif self.update_available:
            self.info_label.setText(self.texts["default"]["update_available"])
        elif self.igame and self.igame.needs_verification:
            self.info_label.setText(self.texts["needs_verification"])
        else:
            self.info_label.setText(" ")  # invisible text, cheap way to always vertical have size in label

    def mousePressEvent(self, e: QMouseEvent):
        # left button
        if e.button() == 1 and not self.game_running:
            if self.igame and self.igame.needs_verification:
                return
            if self.update_available:
                self.launch(skip_version_check=True)
            else:
                self.launch()

        # right
        elif e.button() == 2:
            pass  # self.showMenu(e)

    def sync_finished(self, app_name):
        if not app_name == self.game.app_name:
            return
        super().sync_finished(app_name)
        self.leaveEvent(None)

    def game_finished(self, app_name, error):
        if app_name != self.game.app_name:
            return
        self.game_running = False
        self.leaveEvent(None)

    def game_started(self, app_name):
        if app_name == self.game.app_name:
            self.game_running = True
            self.leaveEvent(None)



import webbrowser

from PyQt5.QtWidgets import QWidget, QVBoxLayout, QMessageBox, QLabel, QPushButton

from rare.shared import LegendaryCoreSingleton, GlobalSignalsSingleton
from rare.utils.misc import icon


class AccountWidget(QWidget):
    def __init__(self):
        super(AccountWidget, self).__init__()
        self.core = LegendaryCoreSingleton()
        self.signals = GlobalSignalsSingleton()

        username = self.core.lgd.userdata.get("display_name")
        if not username:
            username = "Offline"

        self.open_browser = QPushButton(icon("fa.external-link"), self.tr("Account settings"))
        self.open_browser.clicked.connect(
            lambda: webbrowser.open(
                "https://www.epicgames.com/account/personal?productName=epicgames"
            )
        )
        self.logout_button = QPushButton(self.tr("Logout"))
        self.logout_button.clicked.connect(self.logout)

        layout = QVBoxLayout(self)
        layout.addWidget(QLabel(self.tr("Account")))
        layout.addWidget(QLabel(self.tr("Logged in as <b>{}</b>").format(username)))
        layout.addWidget(self.open_browser)
        layout.addWidget(self.logout_button)

    def logout(self):
        reply = QMessageBox.question(
            self.parent().parent(),
            "Message",
            self.tr("Do you really want to logout"),
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No,
        )

        if reply == QMessageBox.Yes:
            self.core.lgd.invalidate_userdata()
            self.signals.exit_app.emit(-133742)  # restart exit code



from hashlib import sha1
from multiprocessing import Queue

# On Windows the monkeypatching of `run_real` below doesn't work like on Linux
# This has the side effect of emitting the UIUpdate in DownloadThread complaining with a TypeError
# So import `legendary.core` and monkeypatch its imported DLManager
import legendary.core
from legendary.core import LegendaryCore as LegendaryCoreReal
from legendary.models.downloading import AnalysisResult
from legendary.models.game import Game, InstalledGame
from legendary.models.manifest import ManifestMeta

from .api_exception import LgndrException, LgndrCoreLogHandler
from .manager import DLManager

legendary.core.DLManager = DLManager


# fmt: off
class LegendaryCore(LegendaryCoreReal):

    def __init__(self, override_config=None, timeout=10.0):
        super(LegendaryCore, self).__init__(override_config=override_config, timeout=timeout)
        self.handler = LgndrCoreLogHandler()
        self.log.addHandler(self.handler)

    # skip_sync defaults to false but since Rare is persistent, skip by default
    # def get_installed_game(self, app_name, skip_sync=True) -> InstalledGame:
    #     return super(LegendaryCore, self).get_installed_game(app_name, skip_sync)

    def prepare_download(self, game: Game, base_game: Game = None, base_path: str = '',
                         status_q: Queue = None, max_shm: int = 0, max_workers: int = 0,
                         force: bool = False, disable_patching: bool = False,
                         game_folder: str = '', override_manifest: str = '',
                         override_old_manifest: str = '', override_base_url: str = '',
                         platform: str = 'Windows', file_prefix_filter: list = None,
                         file_exclude_filter: list = None, file_install_tag: list = None,
                         dl_optimizations: bool = False, dl_timeout: int = 10,
                         repair: bool = False, repair_use_latest: bool = False,
                         disable_delta: bool = False, override_delta_manifest: str = '',
                         egl_guid: str = '', preferred_cdn: str = None,
                         disable_https: bool = False) -> (DLManager, AnalysisResult, ManifestMeta):
        dlm, analysis, igame = super(LegendaryCore, self).prepare_download(
            game=game, base_game=base_game, base_path=base_path,
            status_q=status_q, max_shm=max_shm, max_workers=max_workers,
            force=force, disable_patching=disable_patching,
            game_folder=game_folder, override_manifest=override_manifest,
            override_old_manifest=override_old_manifest, override_base_url=override_base_url,
            platform=platform, file_prefix_filter=file_prefix_filter,
            file_exclude_filter=file_exclude_filter, file_install_tag=file_install_tag,
            dl_optimizations=dl_optimizations, dl_timeout=dl_timeout,
            repair=repair, repair_use_latest=repair_use_latest,
            disable_delta=disable_delta, override_delta_manifest=override_delta_manifest,
            egl_guid=egl_guid, preferred_cdn=preferred_cdn,
            disable_https=disable_https
        )
        # lk: monkeypatch run_real (the method that emits the stats) into DLManager

        # pylint: disable=E1111
        dlm.run_real = DLManager.run_real.__get__(dlm, DLManager)
        # lk: set the queue for reporting statistics back the UI
        dlm.status_queue = Queue()
        # lk: set the queue to send control signals to the DLManager
        # lk: this doesn't exist in the original class, but it is monkeypatched in
        dlm.signals_queue = Queue()
        return dlm, analysis, igame

    def uninstall_game(self, installed_game: InstalledGame, delete_files=True, delete_root_directory=False):
        try:
            super(LegendaryCore, self).uninstall_game(installed_game, delete_files, delete_root_directory)
        except Exception as e:
            raise e
        finally:
            pass

    def egl_import(self, app_name):
        try:
            super(LegendaryCore, self).egl_import(app_name)
        except LgndrException as ret:
            raise ret
        finally:
            pass

    def egl_export(self, app_name):
        try:
            super(LegendaryCore, self).egl_export(app_name)
        except LgndrException as ret:
            raise ret
        finally:
            pass

    def prepare_overlay_install(self, path=None):
        dlm, analysis_result, igame = super(LegendaryCore, self).prepare_overlay_install(path)
        # lk: monkeypatch status_q (the queue for download stats)
        # pylint: disable=E1111
        dlm.run_real = DLManager.run_real.__get__(dlm, DLManager)
        # lk: set the queue for reporting statistics back the UI
        dlm.status_queue = Queue()
        # lk: set the queue to send control signals to the DLManager
        # lk: this doesn't exist in the original class, but it is monkeypatched in
        dlm.signals_queue = Queue()
        return dlm, analysis_result, igame

# fmt: on



import logging
import os
import queue
import subprocess
import time
from typing import Optional, Union, Tuple

from legendary.cli import LegendaryCLI as LegendaryCLIReal
from legendary.models.downloading import AnalysisResult, ConditionCheckResult
from legendary.models.game import Game, InstalledGame, VerifyResult
from legendary.utils.lfs import validate_files
from legendary.utils.selective_dl import get_sdl_appname

from .api_arguments import (
    LgndrInstallGameArgs,
    LgndrImportGameArgs,
    LgndrVerifyGameArgs,
    LgndrUninstallGameArgs,
    LgndrInstallGameRealArgs,
    LgndrInstallGameRealRet,
)
from .api_monkeys import LgndrIndirectStatus, LgndrIndirectLogger
from .core import LegendaryCore
from .manager import DLManager


# fmt: off
class LegendaryCLI(LegendaryCLIReal):

    # noinspection PyMissingConstructor
    def __init__(self, core: LegendaryCore):
        self.core = core
        self.logger = logging.getLogger('Cli')
        self.logging_queue = None
        self.ql = self.setup_threaded_logging()

    def __del__(self):
        self.ql.stop()

    def resolve_aliases(self, name):
        return super(LegendaryCLI, self)._resolve_aliases(name)

    def install_game(self, args: LgndrInstallGameArgs) -> Optional[Tuple[DLManager, AnalysisResult, InstalledGame, Game, bool, Optional[str], ConditionCheckResult]]:
        # Override logger for the local context to use message as part of the indirect return value
        logger = LgndrIndirectLogger(args.indirect_status, self.logger)
        get_boolean_choice = args.get_boolean_choice
        sdl_prompt = args.sdl_prompt

        args.app_name = self._resolve_aliases(args.app_name)
        if self.core.is_installed(args.app_name):
            igame = self.core.get_installed_game(args.app_name)
            args.platform = igame.platform
            if igame.needs_verification and not args.repair_mode:
                logger.info('Game needs to be verified before updating, switching to repair mode...')
                args.repair_mode = True

        repair_file = None
        # Rare: The 'args.no_install' flags is set externally from the InstallDialog
        if args.repair_mode:
            args.repair_mode = True
            args.no_install = args.repair_and_update is False
            repair_file = os.path.join(self.core.lgd.get_tmp_path(), f'{args.app_name}.repair')

        # Rare: Rare is already logged in

        if args.file_prefix or args.file_exclude_prefix:
            args.no_install = True

        # Rare: Rare runs updates on already installed games only

        game = self.core.get_game(args.app_name, update_meta=True, platform=args.platform)

        if not game:
            logger.error(f'Could not find "{args.app_name}" in list of available games, '
                         f'did you type the name correctly?')
            return

        # Rare: Rare checks this before calling 'install_game'

        if args.platform not in game.asset_infos:
            if not args.no_install:
                if self.core.lgd.config.getboolean('Legendary', 'install_platform_fallback', fallback=True):
                    logger.warning(f'App has no asset for platform "{args.platform}", falling back to "Windows".')
                    args.platform = 'Windows'
                else:
                    logger.error(f'No app asset found for platform "{args.platform}", run '
                                 f'"legendary info --platform {args.platform}" and make '
                                 f'sure the app is available for the specified platform.')
                    return
            else:
                logger.warning(f'No asset found for platform "{args.platform}", '
                               f'trying anyway since --no-install is set.')

        if game.is_dlc:
            logger.info('Install candidate is DLC')
            app_name = game.metadata['mainGameItem']['releaseInfo'][0]['appId']
            base_game = self.core.get_game(app_name)
            # check if base_game is actually installed
            if not self.core.is_installed(app_name):
                # download mode doesn't care about whether something's installed
                if not args.no_install:
                    logger.fatal(f'Base game "{app_name}" is not installed!')
                    return
        else:
            base_game = None

        if args.repair_mode:
            if not self.core.is_installed(game.app_name):
                logger.error(f'Game "{game.app_title}" ({game.app_name}) is not installed!')
                return

            if not os.path.exists(repair_file):
                logger.info('Game has not been verified yet.')
                # Rare: we do not want to verify while preparing the download in the InstallDialog
                # Rare: we handle it differently through the GameInfo tab
                logger.error('Game has not been verified yet.')
                return
            else:
                logger.info(f'Using existing repair file: {repair_file}')

        # check if SDL should be disabled
        sdl_enabled = not args.install_tag and not game.is_dlc
        config_tags = self.core.lgd.config.get(game.app_name, 'install_tags', fallback=None)
        config_disable_sdl = self.core.lgd.config.getboolean(game.app_name, 'disable_sdl', fallback=False)
        # remove config flag if SDL is reset
        if config_disable_sdl and args.reset_sdl and not args.disable_sdl:
            self.core.lgd.config.remove_option(game.app_name, 'disable_sdl')
        # if config flag is not yet set, set it and remove previous install tags
        elif not config_disable_sdl and args.disable_sdl:
            logger.info('Clearing install tags from config and disabling SDL for title.')
            if config_tags:
                self.core.lgd.config.remove_option(game.app_name, 'install_tags')
                config_tags = None
            self.core.lgd.config.set(game.app_name, 'disable_sdl', True)
            sdl_enabled = False
        # just disable SDL, but keep config tags that have been manually specified
        elif config_disable_sdl or args.disable_sdl:
            sdl_enabled = False

        if sdl_enabled and ((sdl_name := get_sdl_appname(game.app_name)) is not None):
            if not self.core.is_installed(game.app_name) or config_tags is None or args.reset_sdl:
                sdl_data = self.core.get_sdl_data(sdl_name, platform=args.platform)
                if sdl_data:
                    if args.skip_sdl:
                        args.install_tag = ['']
                        if '__required' in sdl_data:
                            args.install_tag.extend(sdl_data['__required']['tags'])
                    else:
                        args.install_tag = sdl_prompt(sdl_data, game.app_title)
                    self.core.lgd.config.set(game.app_name, 'install_tags', ','.join(args.install_tag))
                else:
                    logger.error(f'Unable to get SDL data for {sdl_name}')
            else:
                args.install_tag = config_tags.split(',')
        elif args.install_tag and not game.is_dlc and not args.no_install:
            config_tags = ','.join(args.install_tag)
            logger.info(f'Saving install tags for "{game.app_name}" to config: {config_tags}')
            self.core.lgd.config.set(game.app_name, 'install_tags', config_tags)
        elif not game.is_dlc:
            if config_tags and args.reset_sdl:
                logger.info('Clearing install tags from config.')
                self.core.lgd.config.remove_option(game.app_name, 'install_tags')
            elif config_tags:
                logger.info(f'Using install tags from config: {config_tags}')
                args.install_tag = config_tags.split(',')

        logger.info(f'Preparing download for "{game.app_title}" ({game.app_name})...')
        # todo use status queue to print progress from CLI
        # This has become a little ridiculous hasn't it?
        dlm, analysis, igame = self.core.prepare_download(game=game, base_game=base_game, base_path=args.base_path,
                                                          force=args.force, max_shm=args.shared_memory,
                                                          max_workers=args.max_workers, game_folder=args.game_folder,
                                                          disable_patching=args.disable_patching,
                                                          override_manifest=args.override_manifest,
                                                          override_old_manifest=args.override_old_manifest,
                                                          override_base_url=args.override_base_url,
                                                          platform=args.platform,
                                                          file_prefix_filter=args.file_prefix,
                                                          file_exclude_filter=args.file_exclude_prefix,
                                                          file_install_tag=args.install_tag,
                                                          dl_optimizations=args.order_opt,
                                                          dl_timeout=args.dl_timeout,
                                                          repair=args.repair_mode,
                                                          repair_use_latest=args.repair_and_update,
                                                          disable_delta=args.disable_delta,
                                                          override_delta_manifest=args.override_delta_manifest,
                                                          preferred_cdn=args.preferred_cdn,
                                                          disable_https=args.disable_https)

        # game is either up-to-date or hasn't changed, so we have nothing to do
        if not analysis.dl_size:
            logger.info('Download size is 0, the game is either already up to date or has not changed. Exiting...')
            self.install_game_cleanup(game, igame, args.repair_mode, repair_file)
            return

        res = self.core.check_installation_conditions(analysis=analysis, install=igame, game=game,
                                                      updating=self.core.is_installed(args.app_name),
                                                      ignore_space_req=args.ignore_space)

        return dlm, analysis, igame, game, args.repair_mode, repair_file, res

    # Rare: This is currently handled in DownloadThread, this is a trial
    def install_game_real(self, args: LgndrInstallGameRealArgs, dlm: DLManager, game: Game, igame: InstalledGame) -> LgndrInstallGameRealRet:
        # Override logger for the local context to use message as part of the indirect return value
        logger = LgndrIndirectLogger(args.indirect_status, self.logger)
        ret = LgndrInstallGameRealRet(game.app_name)

        start_t = time.time()

        try:
            # set up logging stuff (should be moved somewhere else later)
            dlm.logging_queue = self.logging_queue
            dlm.proc_debug = args.dlm_debug

            dlm.start()
            while dlm.is_alive():
                try:
                    args.ui_update(dlm.status_queue.get(timeout=1.0))
                except queue.Empty:
                    pass
                if args.dlm_signals.update:
                    try:
                        dlm.signals_queue.put(args.dlm_signals, block=False, timeout=1.0)
                    except queue.Full:
                        pass
                time.sleep(dlm.update_interval / 10)
            dlm.join()
        except Exception as e:
            end_t = time.time()
            logger.info(f'Installation failed after {end_t - start_t:.02f} seconds.')
            logger.warning(f'The following exception occurred while waiting for the downloader to finish: {e!r}. '
                           f'Try restarting the process, the resume file will be used to start where it failed. '
                           f'If it continues to fail please open an issue on GitHub.')
            ret.ret_code = ret.ReturnCode.ERROR
            ret.message = f"{e!r}"
            return ret
        else:
            end_t = time.time()
            if args.dlm_signals.kill is True:
                logger.info(f"Download stopped after {end_t - start_t:.02f} seconds.")
                ret.exit_code = ret.ReturnCode.STOPPED
                return ret
            logger.info(f"Download finished in {end_t - start_t:.02f} seconds.")
            if not args.no_install:
                # Allow setting savegame directory at install time so sync-saves will work immediately
                if (game.supports_cloud_saves or game.supports_mac_cloud_saves) and args.save_path:
                    igame.save_path = args.save_path

                postinstall = self.core.install_game(igame)
                if postinstall:
                    self._handle_postinstall(postinstall, igame, skip_prereqs=args.yes, choice=args.install_prereqs)

                dlcs = self.core.get_dlc_for_game(game.app_name)
                if dlcs and not args.skip_dlcs:
                    for dlc in dlcs:
                        ret.dlcs.append(
                            {
                                "app_name": dlc.app_name,
                                "app_title": dlc.app_title,
                                "app_version": dlc.app_version(args.platform)
                            }
                        )

                    # Rare: We do not install DLCs automatically, we offer to do so through our downloads tab

                if (game.supports_cloud_saves or game.supports_mac_cloud_saves) and not game.is_dlc:
                    # todo option to automatically download saves after the installation
                    #  args does not have the required attributes for sync_saves in here,
                    #  not sure how to solve that elegantly.
                    logger.info(f'This game supports cloud saves, syncing is handled by the "sync-saves" command. '
                                f'To download saves for this game run "legendary sync-saves {args.app_name}"')
                    ret.sync_saves = True

                # show tip again after installation finishes so users hopefully actually see it
                if tip_url := self.core.get_game_tip(igame.app_name):
                    ret.tip_url = tip_url

            self.install_game_cleanup(game, igame, args.repair_mode, args.repair_file)

            logger.info(f'Finished installation process in {end_t - start_t:.02f} seconds.')

            return ret

    def install_game_cleanup(self, game: Game, igame: InstalledGame, repair_mode: bool = False, repair_file: str = '') -> None:
        # Override logger for the local context to use message as part of the indirect return value
        logger = LgndrIndirectLogger(LgndrIndirectStatus(), self.logger)

        old_igame = self.core.get_installed_game(game.app_name)
        if old_igame and repair_mode and os.path.exists(repair_file):
            if old_igame.needs_verification:
                old_igame.needs_verification = False
                self.core.install_game(old_igame)

            logger.debug('Removing repair file.')
            os.remove(repair_file)

        # check if install tags have changed, if they did; try deleting files that are no longer required.
        if old_igame and old_igame.install_tags != igame.install_tags:
            old_igame.install_tags = igame.install_tags
            logger.info('Deleting now untagged files.')
            self.core.uninstall_tag(old_igame)
            self.core.install_game(old_igame)

    def _handle_postinstall(self, postinstall, igame, skip_prereqs=False, choice=True):
        # Override logger for the local context to use message as part of the indirect return value
        logger = LgndrIndirectLogger(LgndrIndirectStatus(), self.logger)
        # noinspection PyShadowingBuiltins
        def print(x): self.logger.info(x) if x else None
        # noinspection PyShadowingBuiltins
        def input(x): return 'y' if choice else 'n'

        print('\nThis game lists the following prerequisites to be installed:')
        print(f'- {postinstall["name"]}: {" ".join((postinstall["path"], postinstall["args"]))}')
        print('')

        if os.name == 'nt':
            if skip_prereqs:
                c = 'n'  # we don't want to launch anything, just silent install.
            else:
                choice = input('Do you wish to install the prerequisites? ([y]es, [n]o, [i]gnore): ')
                c = choice.lower()[0]
                print('')

            if c == 'i':  # just set it to installed
                logger.info('Marking prerequisites as installed...')
                self.core.prereq_installed(igame.app_name)
            elif c == 'y':  # set to installed and launch installation
                logger.info('Launching prerequisite executable..')
                self.core.prereq_installed(igame.app_name)
                req_path, req_exec = os.path.split(postinstall['path'])
                work_dir = os.path.join(igame.install_path, req_path)
                fullpath = os.path.join(work_dir, req_exec)
                try:
                    p = subprocess.Popen([fullpath, postinstall['args']], cwd=work_dir, shell=True)
                    p.wait()
                except Exception as e:
                    logger.error(f'Failed to run prereq executable with: {e!r}')
        else:
            logger.info('Automatic installation not available on Linux.')

    def uninstall_game(self, args: LgndrUninstallGameArgs) -> None:
        # Override logger for the local context to use message as part of the indirect return value
        logger = LgndrIndirectLogger(args.indirect_status, self.logger, logging.WARNING)
        get_boolean_choice = args.get_boolean_choice

        args.app_name = self._resolve_aliases(args.app_name)
        igame = self.core.get_installed_game(args.app_name)
        if not igame:
            logger.error(f'Game {args.app_name} not installed, cannot uninstall!')
            return

        if not args.yes:
            if not get_boolean_choice(f'Do you wish to uninstall "{igame.title}"?', default=False):
                return

        try:
            if not igame.is_dlc:
                # Remove DLC first so directory is empty when game uninstall runs
                dlcs = self.core.get_dlc_for_game(igame.app_name)
                for dlc in dlcs:
                    if (idlc := self.core.get_installed_game(dlc.app_name)) is not None:
                        logger.info(f'Uninstalling DLC "{dlc.app_name}"...')
                        self.core.uninstall_game(idlc, delete_files=not args.keep_files)

            logger.info(f'Removing "{igame.title}" from "{igame.install_path}"...')
            self.core.uninstall_game(igame, delete_files=not args.keep_files,
                                     delete_root_directory=not igame.is_dlc)
            logger.info('Game has been uninstalled.')
            return
        except Exception as e:
            logger.warning(f'Removing game failed: {e!r}, please remove {igame.install_path} manually.')
            return

    def verify_game(self, args: Union[LgndrVerifyGameArgs, LgndrInstallGameArgs], print_command=True, repair_mode=False, repair_online=False) -> Optional[Tuple[int, int]]:
        # Override logger for the local context to use message as part of the indirect return value
        logger = LgndrIndirectLogger(args.indirect_status, self.logger)

        args.app_name = self._resolve_aliases(args.app_name)
        if not self.core.is_installed(args.app_name):
            logger.error(f'Game "{args.app_name}" is not installed')
            return

        logger.info(f'Loading installed manifest for "{args.app_name}"')
        igame = self.core.get_installed_game(args.app_name)
        if not os.path.exists(igame.install_path):
            logger.error(f'Install path "{igame.install_path}" does not exist, make sure all necessary mounts '
                         f'are available. If you previously deleted the game folder without uninstalling, run '
                         f'"legendary uninstall -y {igame.app_name}" and reinstall from scratch.')
            return

        manifest_data, _ = self.core.get_installed_manifest(args.app_name)
        if manifest_data is None:
            if repair_mode:
                if not repair_online:
                    logger.critical('No manifest could be loaded, the manifest file may be missing!')
                    raise ValueError('Local manifest is missing')

                logger.warning('No manifest could be loaded, the file may be missing. Downloading the latest manifest.')
                game = self.core.get_game(args.app_name, platform=igame.platform)
                manifest_data, _ = self.core.get_cdn_manifest(game, igame.platform)
                # Rare: Save the manifest if we downloaded it because it was missing
                self.core.lgd.save_manifest(game.app_name, manifest_data,
                                            version=self.core.load_manifest(manifest_data).meta.build_version,
                                            platform=igame.platform)
            else:
                logger.critical(f'Manifest appears to be missing! To repair, run "legendary repair '
                                f'{args.app_name} --repair-and-update", this will however redownload all files '
                                f'that do not match the latest manifest in their entirety.')
                return

        manifest = self.core.load_manifest(manifest_data)

        files = sorted(manifest.file_manifest_list.elements,
                       key=lambda a: a.filename.lower())

        # build list of hashes
        if config_tags := self.core.lgd.config.get(args.app_name, 'install_tags', fallback=None):
            install_tags = set(i.strip() for i in config_tags.split(','))
            file_list = [
                (f.filename, f.sha_hash.hex())
                for f in files
                if any(it in install_tags for it in f.install_tags) or not f.install_tags
            ]
        else:
            file_list = [(f.filename, f.sha_hash.hex()) for f in files]

        total = len(file_list)
        total_size = sum(manifest.file_manifest_list.get_file_by_path(fm[0]).file_size
                         for fm in file_list)
        num = processed = last_processed = 0
        speed = 0.0
        percentage = 0.0
        failed = []
        missing = []

        last_update = time.time()

        logger.info(f'Verifying "{igame.title}" version "{manifest.meta.build_version}"')
        repair_file = []
        for result, path, result_hash, bytes_read in validate_files(igame.install_path, file_list):
            processed += bytes_read
            percentage = (processed / total_size) * 100.0
            num += 1

            if (delta := ((current_time := time.time()) - last_update)) > 1:
                last_update = current_time
                speed = (processed - last_processed) / 1024 / 1024 / delta
                last_processed = processed

            if args.verify_stdout:
                args.verify_stdout(num, total, percentage, speed)

            if result == VerifyResult.HASH_MATCH:
                repair_file.append(f'{result_hash}:{path}')
                continue
            elif result == VerifyResult.HASH_MISMATCH:
                logger.error(f'File does not match hash: "{path}"')
                repair_file.append(f'{result_hash}:{path}')
                failed.append(path)
            elif result == VerifyResult.FILE_MISSING:
                logger.error(f'File is missing: "{path}"')
                missing.append(path)
            else:
                logger.error(f'Other failure (see log), treating file as missing: "{path}"')
                missing.append(path)

        if args.verify_stdout:
            args.verify_stdout(num, total, percentage, speed)

        # always write repair file, even if all match
        if repair_file:
            repair_filename = os.path.join(self.core.lgd.get_tmp_path(), f'{args.app_name}.repair')
            with open(repair_filename, 'w', encoding='utf-8') as f:
                f.write('\n'.join(repair_file))
            logger.debug(f'Written repair file to "{repair_filename}"')

        if not missing and not failed:
            logger.info('Verification finished successfully.')
            return 0, 0
        else:
            logger.error(f'Verification failed, {len(failed)} file(s) corrupted, {len(missing)} file(s) are missing.')
            if print_command:
                logger.info(f'Run "legendary repair {args.app_name}" to repair your game installation.')
            return len(failed), len(missing)

    def import_game(self, args: LgndrImportGameArgs) -> None:
        # Override logger for the local context to use message as part of the indirect return value
        logger = LgndrIndirectLogger(args.indirect_status, self.logger)
        get_boolean_choice = args.get_boolean_choice

        # make sure path is absolute
        args.app_path = os.path.abspath(args.app_path)
        args.app_name = self._resolve_aliases(args.app_name)

        if not os.path.exists(args.app_path):
            logger.error(f'Specified path "{args.app_path}" does not exist!')
            return

        if self.core.is_installed(args.app_name):
            logger.error('Game is already installed!')
            return

        if not self.core.login():
            logger.error('Log in failed!')
            return

        # do some basic checks
        game = self.core.get_game(args.app_name, update_meta=True, platform=args.platform)
        if not game:
            logger.fatal(f'Did not find game "{args.app_name}" on account.')
            return

        if game.is_dlc:
            release_info = game.metadata.get('mainGameItem', {}).get('releaseInfo')
            if release_info:
                main_game_appname = release_info[0]['appId']
                main_game_title = game.metadata['mainGameItem']['title']
                if not self.core.is_installed(main_game_appname):
                    logger.error(f'Import candidate is DLC but base game "{main_game_title}" '
                                 f'(App name: "{main_game_appname}") is not installed!')
                    return
            else:
                logger.fatal(f'Unable to get base game information for DLC, cannot continue.')
                return

        # get everything needed for import from core, then run additional checks.
        manifest, igame = self.core.import_game(game, args.app_path, platform=args.platform)
        exe_path = os.path.join(args.app_path, manifest.meta.launch_exe.lstrip('/'))
        # check if most files at least exist or if user might have specified the wrong directory
        total = len(manifest.file_manifest_list.elements)
        found = sum(os.path.exists(os.path.join(args.app_path, f.filename))
                    for f in manifest.file_manifest_list.elements)
        ratio = found / total

        if not found:
            logger.error(f'No files belonging to {"DLC" if game.is_dlc else "Game"} "{game.app_title}" '
                         f'({game.app_name}) found in the specified location, please verify that the path is correct.')
            if not game.is_dlc:
                # check if game folder is in path, suggest alternative
                folder = game.metadata.get('customAttributes', {}).get('FolderName', {}).get('value', game.app_name)
                if folder and folder not in args.app_path:
                    new_path = os.path.join(args.app_path, folder)
                    logger.info(f'Did you mean "{new_path}"?')
            return

        if not game.is_dlc and not os.path.exists(exe_path) and not args.disable_check:
            logger.error(f'Game executable could not be found at "{exe_path}", '
                         f'please verify that the specified path is correct.')
            return

        if ratio < 0.95:
            logger.warning('Some files are missing from the game installation, install may not '
                           'match latest Epic Games Store version or might be corrupted.')
        else:
            logger.info(f'{"DLC" if game.is_dlc else "Game"} install appears to be complete.')

        self.core.install_game(igame)
        if igame.needs_verification:
            logger.info(f'NOTE: The {"DLC" if game.is_dlc else "Game"} installation will have to be '
                        f'verified before it can be updated with legendary.')
            logger.info(f'Run "legendary repair {args.app_name}" to do so.')
        else:
            logger.info(f'Installation had Epic Games Launcher metadata for version "{igame.version}", '
                        f'verification will not be required.')

        # check for importable DLC
        if not args.skip_dlcs:
            dlcs = self.core.get_dlc_for_game(game.app_name)
            if dlcs:
                logger.info(f'Found {len(dlcs)} items of DLC that could be imported.')
                import_dlc = True
                if not args.yes and not args.with_dlcs:
                    if not get_boolean_choice(f'Do you wish to automatically attempt to import all DLCs?'):
                        import_dlc = False

                if import_dlc:
                    for dlc in dlcs:
                        args.app_name = dlc.app_name
                        self.import_game(args)

        logger.info(f'{"DLC" if game.is_dlc else "Game"} "{game.app_title}" has been imported.')
        return

    def move(self, args):
        # Override logger for the local context to use message as part of the indirect return value
        logger = LgndrIndirectLogger(args.indirect_status, self.logger)

        app_name = self._resolve_aliases(args.app_name)
        igame = self.core.get_installed_game(app_name, skip_sync=True)
        if not igame:
            logger.error(f'No installed game found for "{app_name}"')
            return

        old_base, game_folder = os.path.split(igame.install_path.replace('\\', '/'))
        new_path = os.path.join(args.new_path, game_folder)
        logger.info(f'Moving "{game_folder}" from "{old_base}" to "{args.new_path}"')

        if not args.skip_move:
            try:
                if not os.path.exists(args.new_path):
                    os.makedirs(args.new_path)

                os.rename(igame.install_path, new_path)
            except Exception as e:
                # pylint: disable=E1101
                if isinstance(e, OSError) and e.errno == 18:
                    logger.error(f'Moving to a different drive is not supported. Move the folder manually to '
                                 f'"{new_path}" and run "legendary move {app_name} "{args.new_path}" --skip-move"')
                elif isinstance(e, FileExistsError):
                    logger.error(f'The target path already contains a folder called "{game_folder}", '
                                 f'please remove or rename it first.')
                else:
                    logger.error(f'Moving failed with unknown error {e!r}.')
                    logger.info(f'Try moving the folder manually to "{new_path}" and running '
                                f'"legendary move {app_name} "{args.new_path}" --skip-move"')
                return
        else:
            logger.info(f'Not moving, just rewriting legendary metadata...')

        igame.install_path = new_path
        self.core.install_game(igame)
        logger.info('Finished.')

# fmt: on



import platform
from logging import getLogger

from PyQt5.QtCore import QSettings, Qt
from PyQt5.QtWidgets import (
    QWidget,
    QLabel
)

from rare.components.tabs.settings.widgets.env_vars import EnvVars
from rare.components.tabs.settings.widgets.linux import LinuxSettings
from rare.components.tabs.settings.widgets.proton import ProtonSettings
from rare.components.tabs.settings.widgets.wrapper import WrapperSettings
from rare.shared import LegendaryCoreSingleton
from rare.ui.components.tabs.games.game_info.game_settings import Ui_GameSettings
from rare.utils import config_helper

logger = getLogger("GameSettings")


class DefaultGameSettings(QWidget, Ui_GameSettings):
    # variable to no update when changing game
    change = False
    app_name: str

    def __init__(self, is_default, parent=None):
        super(DefaultGameSettings, self).__init__(parent=parent)
        self.setupUi(self)
        self.core = LegendaryCoreSingleton()
        self.settings = QSettings()

        self.wrapper_settings = WrapperSettings()

        self.launch_settings_group.layout().addRow(
            QLabel("Wrapper"), self.wrapper_settings
        )

        if platform.system() != "Windows":
            self.linux_settings = LinuxAppSettings()
            self.proton_settings = ProtonSettings(self.linux_settings, self.wrapper_settings)
            self.proton_layout.addWidget(self.proton_settings)

            # FIXME: Remove the spacerItem and margins from the linux settings
            # FIXME: This should be handled differently at soem point in the future
            self.linux_settings.layout().setContentsMargins(0, 0, 0, 0)
            for item in [
                self.linux_settings.layout().itemAt(idx)
                for idx in range(self.linux_settings.layout().count())
            ]:
                if item.spacerItem():
                    self.linux_settings.layout().removeItem(item)
                    del item
            # FIXME: End of FIXME
            self.linux_settings_layout.addWidget(self.linux_settings)
            self.linux_settings_layout.setAlignment(Qt.AlignTop)

            self.game_settings_layout.setAlignment(Qt.AlignTop)

            self.linux_settings.mangohud.set_wrapper_activated.connect(
                lambda active: self.wrapper_settings.add_wrapper("mangohud")
                if active else self.wrapper_settings.delete_wrapper("mangohud"))

        else:
            self.linux_settings_widget.setVisible(False)

        self.env_vars = EnvVars(self)
        self.game_settings_layout.addWidget(self.env_vars)

        if is_default:
            for i in range(4):  # remove some entries which are not supported by default
                self.launch_settings_layout.removeRow(0)

            self.cloud_group.deleteLater()
            self.load_settings("default")

    def save_line_edit(self, option, value):
        if value:
            config_helper.add_option(self.game.app_name, option, value)
        else:
            config_helper.remove_option(self.game.app_name, option)
        config_helper.save_config()

        if option == "wine_prefix":
            if self.game.supports_cloud_saves:
                self.compute_save_path()

    def update_combobox(self, i, option):
        if self.change:
            # remove section
            if i:
                if i == 1:
                    config_helper.add_option(self.game.app_name, option, "true")
                if i == 2:
                    config_helper.add_option(self.game.app_name, option, "false")
            else:
                config_helper.remove_option(self.game.app_name, option)
            config_helper.save_config()

    def load_settings(self, app_name):
        self.app_name = app_name
        self.wrapper_settings.load_settings(app_name)
        if platform.system() != "Windows":
            self.linux_settings.update_game(app_name)
            proton = self.wrapper_settings.wrappers.get("proton", "")
            if proton:
                proton = proton.text
            self.proton_settings.load_settings(app_name, proton)
            if proton:
                self.linux_settings.wine_groupbox.setEnabled(False)
            else:
                self.linux_settings.wine_groupbox.setEnabled(True)
        self.env_vars.update_game(app_name)


class LinuxAppSettings(LinuxSettings):
    def __init__(self):
        super(LinuxAppSettings, self).__init__()

    def update_game(self, app_name):
        self.name = app_name
        self.wine_prefix.setText(self.load_prefix())
        self.wine_exec.setText(self.load_setting(self.name, "wine_executable"))

        self.dxvk.load_settings(self.name)

        self.mangohud.load_settings(self.name)



from logging import getLogger

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QVBoxLayout, QWidget, QHBoxLayout
from legendary.core import LegendaryCore
from legendary.models.game import Game

from rare.components.tabs.games.game_widgets.base_uninstalled_widget import (
    BaseUninstalledWidget,
)
from rare.shared.image_manager import ImageSize
from rare.widgets.elide_label import ElideLabel

logger = getLogger("Uninstalled")


class UninstalledIconWidget(BaseUninstalledWidget):
    def __init__(self, game: Game, core: LegendaryCore, pixmap):
        super(UninstalledIconWidget, self).__init__(game, core, pixmap)
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setObjectName(type(self).__name__)
        layout.addWidget(self.image)

        miniwidget = QWidget(self)
        miniwidget.setFixedWidth(ImageSize.Display.size.width())
        minilayout = QHBoxLayout()
        minilayout.setContentsMargins(0, 0, 0, 0)
        minilayout.setSpacing(0)
        miniwidget.setLayout(minilayout)

        self.title_label = ElideLabel(f"<b>{game.app_title}</b>", parent=miniwidget)
        self.title_label.setAlignment(Qt.AlignTop)
        self.title_label.setObjectName("game_widget")
        minilayout.addWidget(self.title_label, stretch=2)

        minilayout.setAlignment(Qt.AlignTop)
        layout.addWidget(miniwidget)

        self.info_label = ElideLabel(" ", parent=self)
        self.info_label.setFixedWidth(ImageSize.Display.size.width())
        self.leaveEvent(None)
        self.info_label.setObjectName("info_label")
        layout.addWidget(self.info_label)

        self.setLayout(layout)

    def mousePressEvent(self, e) -> None:
        # left button
        if e.button() == 1 and not self.installing:
            self.install()

        # right
        elif e.button() == 2:
            pass  # self.showMenu(e)

    def enterEvent(self, e):
        if not self.installing:
            self.info_label.setText(self.tr("Game Info"))
        else:
            self.info_label.setText(self.tr("Installation running"))

    def leaveEvent(self, e):
        if self.installing:
            self.info_label.setText("Installation...")
        else:
            self.info_label.setText(" ")  # invisible text, cheap way to always have vertical size in label



from logging import getLogger

from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtWidgets import QFrame, QAction
from legendary.models.game import Game

from rare.shared import ImageManagerSingleton
from rare.shared.image_manager import ImageSize
from rare.widgets.image_widget import ImageWidget

logger = getLogger("Uninstalled")


class BaseUninstalledWidget(QFrame):
    show_uninstalled_info = pyqtSignal(Game)

    def __init__(self, game, core, pixmap):
        super(BaseUninstalledWidget, self).__init__()
        self.image_manager = ImageManagerSingleton()

        self.game = game
        if self.game.app_title == "Unreal Engine":
            self.game.app_title = f"{self.game.app_title} {self.game.app_name.split('_')[-1]}"

        self.core = core
        self.image = ImageWidget(self)
        self.image.setFixedSize(ImageSize.Display)
        self.image.setPixmap(pixmap)
        self.installing = False
        self.setContextMenuPolicy(Qt.ActionsContextMenu)

        reload_image = QAction(self.tr("Reload Image"), self)
        reload_image.triggered.connect(self.reload_image)
        self.addAction(reload_image)

    def reload_image(self):
        self.image_manager.download_image_blocking(self.game, force=True)
        pm = self.image_manager.get_pixmap(self.game.app_name, color=False)
        self.image.setPixmap(pm)

    def install(self):
        self.show_uninstalled_info.emit(self.game)

    # From RareGame, added from sorting to work
    @property
    def is_non_asset(self) -> bool:
        """!
        @brief Property to report if a Game doesn't have assets

        Typically, games have assets, however some games that require
        other launchers do not have them. Rare treats these games as installed
        offering to execute their launcher.

        @return bool If the game doesn't have assets
        """
        return not self.game.asset_infos

    @property
    def is_installed(self) -> bool:
        """!
        @brief Property to report if a game is installed

        This returns True if InstalledGame data have been loaded for the game
        or if the game is a game without assets, for example an Origin game.

        @return bool If the game should be considered installed
        """
        return False or self.is_non_asset

    @property
    def app_title(self) -> str:
        return self.game.app_title
